use gdk::gio;
use gdk::prelude::{Cast, ListModelExt};
use gdk::subclass::prelude::ObjectSubclassIsExt;
use glib::object::ObjectExt;
use glib::{clone, Object};

use super::Video;

// TODO: Section model?
gtk::glib::wrapper! {
    pub struct WatchLaterStore(ObjectSubclass<imp::WatchLaterStore>)
        @implements gio::ListModel, gtk::SelectionModel;
}

impl Default for WatchLaterStore {
    fn default() -> Self {
        Object::builder().build()
    }
}

impl WatchLaterStore {
    /// Toggle a video in the list.
    ///
    /// Returns `true` if the video was added to the list, and `false` if it was removed.
    pub fn toggle(&self, video: Video) -> bool {
        let mut videos = self.imp().videos.borrow_mut();

        let index = videos.binary_search_by(|v| {
            v.uploaded()
                .cmp(&video.uploaded())
                .then_with(|| v.id().id.cmp(&video.id().id))
                .then_with(|| v.platform().cmp(&video.platform()))
                .reverse()
        });
        match index {
            Ok(index) => {
                // Remove any installed signal handler.
                if let Some(handler) = self.imp().signal_handlers.borrow_mut().remove(&video.id()) {
                    video.disconnect(handler);
                } else {
                    log::warn!("Removing a video from the watch later list, which does not have a handler installed. This should not happen.");
                }

                videos.remove(index);
                drop(videos);
                self.items_changed(index as u32, 1, 0);
                self.notify_is_empty();
                false
            }
            Err(index) => {
                // As the uploaded time can be inaccurate in the case of YouTube (they only say e.g. "1 day ago"), the uploaded date can vary over time in the watch later list.
                // Add a handler to reinsert the element in the watch later list when that happens.
                let handler = video.connect_uploaded_unix_notify(clone!(
                    #[strong(rename_to=s)]
                    self,
                    move |video| {
                        log::trace!("Video uploaded changed; reinserting into watch later list");
                        s.toggle(video.clone());
                        s.toggle(video.clone());
                    }
                ));
                self.imp()
                    .signal_handlers
                    .borrow_mut()
                    .insert(video.id(), handler);

                videos.insert(index, video);
                drop(videos);
                self.items_changed(index as u32, 0, 1);
                self.notify_is_empty();
                true
            }
        }
    }

    fn items_changed(&self, position: u32, removed: u32, added: u32) {
        self.upcast_ref::<gio::ListModel>()
            .items_changed(position, removed, added);
    }
}

mod imp {
    use gdk::gio;
    use gdk::prelude::Cast;
    use glib::subclass::types::ObjectSubclassExt;
    use glib::types::StaticType;
    use glib::SignalHandlerId;
    use gtk::glib;
    use gtk::prelude::SelectionModelExt;
    use gtk::subclass::selection_model::SelectionModelImpl;
    use pcore::VideoId;
    use std::cell::RefCell;
    use std::collections::HashMap;

    use gdk::subclass::prelude::{DerivedObjectProperties, ListModelImpl};
    use gdk::{
        glib::Properties,
        prelude::ObjectExt,
        subclass::prelude::{ObjectImpl, ObjectSubclass},
    };

    use crate::backend::Video;

    #[derive(Default, Properties)]
    #[properties(wrapper_type = super::WatchLaterStore)]
    pub struct WatchLaterStore {
        #[property(name = "is-empty", type = bool, get = |s: &Self| s.videos.borrow().is_empty())]
        pub(super) videos: RefCell<Vec<Video>>,

        pub(super) signal_handlers: RefCell<HashMap<VideoId, SignalHandlerId>>,

        #[property(get, set = Self::set_selected, nullable)]
        selected: RefCell<Option<Video>>,
    }

    impl WatchLaterStore {
        fn set_selected(&self, selected: Option<Video>) {
            self.selected.replace(selected);
            // Don't know where exactly it changed; emit that it changed everywhere.
            let num_videos = self.videos.borrow().len() as u32;
            if num_videos > 0 {
                self.obj().selection_changed(0, num_videos);
            }
        }
    }

    #[glib::object_subclass]
    impl ObjectSubclass for WatchLaterStore {
        const NAME: &'static str = "TFWatchLaterStore";
        type Type = super::WatchLaterStore;
        type Interfaces = (gio::ListModel, gtk::SelectionModel);
    }

    #[glib::derived_properties]
    impl ObjectImpl for WatchLaterStore {}

    impl ListModelImpl for WatchLaterStore {
        fn item_type(&self) -> glib::Type {
            Video::static_type()
        }

        fn n_items(&self) -> u32 {
            self.videos.borrow().len().try_into().unwrap_or_default()
        }

        fn item(&self, position: u32) -> Option<glib::Object> {
            let list = self.videos.borrow();

            list.get(position as usize)
                .map(|o| o.clone().upcast::<glib::Object>())
        }
    }

    impl SelectionModelImpl for WatchLaterStore {
        fn is_selected(&self, position: u32) -> bool {
            let list = self.videos.borrow();
            let selection = self.selected.borrow();
            let Some(selection) = selection.as_ref() else {
                return false;
            };
            list.get(position as usize)
                .is_some_and(|v| v.id() == selection.id())
        }

        fn selection_in_range(&self, position: u32, n_items: u32) -> gtk::Bitset {
            let result = gtk::Bitset::new_range(position, n_items);
            let list = self.videos.borrow();
            let selection = self.selected.borrow();
            let Some(selection) = selection.as_ref() else {
                return result;
            };

            let index: Option<u32> = list
                .iter()
                .position(|v| v.id() == selection.id())
                .and_then(|v| v.try_into().ok());

            if let Some(index) = index {
                if position <= index && index < position + n_items {
                    result.add(position);
                }
            }
            result
        }
    }
}
